<!DOCTYPE html>
<html lang="en">
<head>
  <%= HeadTemplate.new("") %>
  <meta id="repository-name" content="<%= repository_name %>">
  <title>README - <%= repository_name %></title>
  <script type="text/javascript">
  CrystalDoc.base_path = "";
  </script>
</head>
<body>

<%= SidebarTemplate.new(repository_name, types, nil) %>

<div class="main-content">
<%= body %>
</div>
</body>
</html>
